#include "ESP32BleGamepad.h"
#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEServer.h>
#include <BLEHIDDevice.h>
#include <HIDTypes.h>

#define GAMEPAD_REPORT_ID 0x03
#define GAMEPAD_STATE_SIZE 7
// 報告描述器資料
static const uint8_t _hidReportDescriptor[] = {
	// 遊戲控制器（Gamepad）
	0x05, 0x01, // USAGE_PAGE (Generic Desktop)
	0x09, 0x05, // USAGE (Gamepad)
	0xa1, 0x01, // COLLECTION (Application)
	0x85, 0x03, //   REPORT_ID (3)

	// 14個按鍵
	0x05, 0x09, //   USAGE_PAGE (Button)
	0x19, 0x01, //   USAGE_MINIMUM (範圍最小值：1)
	0x29, 0x0E, //   USAGE_MAXIMUM (範圍最大值：14)
	0x15, 0x00, //   LOGICAL_MINIMUM (0)
	0x25, 0x01, //   LOGICAL_MAXIMUM (1)
	0x75, 0x01, //   REPORT_SIZE (1)
	0x95, 0x0E, //   REPORT_COUNT (14個按鍵)
	0x81, 0x02, //   INPUT (Data,Var,Abs)
	// 補上兩個空白
	0x95, 0x01, //   REPORT_COUNT (1)
	0x75, 0x02, //   REPORT_SIZE (2個位元)
	0x81, 0x01, //   INPUT (保留)

	// 8方向十字鍵（HAT）
	0x05, 0x01,		  //   USAGE_PAGE (General Desktop)，通用桌上型
	0x09, 0x39,		  //   USAGE (Hat Switch)，帽子開關（十字鍵）
	0x15, 0x00,		  //   LOGICAL_MINIMUM (0)，邏輯最小值
	0x25, 0x07,		  //   LOGICAL_MAXIMUM (7)，邏輯最大值
	0x35, 0x00,		  //   PHYSICAL_MINIMUM (0)，實體最小值
	0x46, 0x3B, 0x01, //   PHYSICAL_MAXIMUM (315) ，實體最大值
	0x65, 0x14,		  //   UNIT (Eng Rot:Angular Pos)，單位：英制角度
	0x75, 0x04,		  //   REPORT_SIZE (4)，佔4個位元
	0x95, 0x01,		  //   REPORT_COUNT (1)
	0x81, 0x02,		  //   INPUT (Data,Var,Abs)，絕對可變資料
					  // 填補4個空白位元
	0x95, 0x01,		  // REPORT_COUNT (1)
	0x75, 0x04,		  // REPORT_SIZE (4)，共4個位元
	0x81, 0x01,		  // INPUT

	// X, Y和Z軸
	0x15, 0x00,		  //   LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00, //   LOGICAL_MAXIMUM (255)
	0x75, 0x08,		  //   REPORT_SIZE (8)
	0x09, 0x01,		  //   USAGE (Pointer)，游標
	0xA1, 0x00,		  //   COLLECTION (Physical)，游標的座標資料集合
	0x09, 0x30,		  //     USAGE (x)
	0x09, 0x31,		  //     USAGE (y)
	0x09, 0x32,		  //     USAGE (z)
	0x09, 0x35,		  //     USAGE (rz)
	0x95, 0x04,		  //     REPORT_COUNT (4)
	0x81, 0x02,		  //     INPUT (Data,Var,Abs)
	0xc0,			  //   END_COLLECTION

	0xc0 // END_COLLECTION
};

BleGamepad::BleGamepad(std::string deviceName, std::string deviceManufacturer, uint8_t batteryLevel)
{
	// 初始化類比搖桿、十字鍵和按鍵的狀態
	xAxis = 0;
	yAxis = 0;
	zAxis = 0;
	zAxisRotation = 0;
	buttons = 0;
	hatSwitch = -1;

	this->deviceName = deviceName;						// 裝置名稱字串
	this->deviceManufacturer = deviceManufacturer;		// 製造商名稱字串
	this->batteryLevel = batteryLevel;					// 電量準位
	this->connectionStatus = new BleConnectionStatus(); // 連線狀態物件
}

void BleGamepad::begin(bool initAutoSendState)
{
	autoSendState = initAutoSendState;
	// 建立FreeRTOS任務
	xTaskCreate(this->taskServer, "server", 20000, (void *)this, 5, NULL);
}
// FreeRTOS任務函式
void BleGamepad::taskServer(void *pvParameter)
{
	// 從任務參數取得遊戲控制器物件（BleGamepadInstance）
	BleGamepad *BleGamepadInstance = (BleGamepad *)pvParameter;
	// 設定裝置名稱
	BLEDevice::init(BleGamepadInstance->deviceName);
	// 建立藍牙裝置伺服器
	BLEServer *pServer = BLEDevice::createServer();
	// 設定連線狀態回呼
	pServer->setCallbacks(BleGamepadInstance->connectionStatus);
	// 建立HID（人機介面裝置）物件
	BleGamepadInstance->hid = new BLEHIDDevice(pServer);
	// 設定「輸入」類型裝置的報告ID
	BleGamepadInstance->inputGamepad = BleGamepadInstance->hid->inputReport(GAMEPAD_REPORT_ID);
	// 替連線狀態設定偵聽的對象
	BleGamepadInstance->connectionStatus->inputGamepad = BleGamepadInstance->inputGamepad;
	// 設定製造商資訊（UTF-8編碼字串的自訂廠商名稱）
	BleGamepadInstance->hid->manufacturer()->setValue(BleGamepadInstance->deviceManufacturer);
	// 設定裝置的PID和VID，參閱下文說明。
	BleGamepadInstance->hid->pnp(0x02, 0x7e05, 0x0920, 0x0110);
	BleGamepadInstance->hid->hidInfo(0x00, 0x01);
	// 設定連線認證模式，“ESP_LE_AUTH_BOND”代表「綁定」，
	// 也就是儲存長期密鑰，日後連線不必再配對，
	// 相關說明請參閱《超圖解ESP32深度實作》15-17頁。
	BLESecurity *pSecurity = new BLESecurity();
	pSecurity->setAuthenticationMode(ESP_LE_AUTH_BOND);
	// 設定HID報告描述器
	BleGamepadInstance->hid->reportMap((uint8_t *)_hidReportDescriptor, sizeof(_hidReportDescriptor));
	// 啟動服務
	BleGamepadInstance->hid->startServices();
	// 設定啟動服務的回呼處理函式
	BleGamepadInstance->onStarted(pServer);
	// 設定廣播訊息（廣告封包）
	BLEAdvertising *pAdvertising = pServer->getAdvertising();
	// 設定HID裝置的圖示
	pAdvertising->setAppearance(HID_GAMEPAD);
	pAdvertising->addServiceUUID(BleGamepadInstance->hid->hidService()->getUUID());
	// 開始廣播
	pAdvertising->start();
	// 設定電量準位
	BleGamepadInstance->hid->setBatteryLevel(BleGamepadInstance->batteryLevel);
	// 無限期等待
	vTaskDelay(portMAX_DELAY);
}

void BleGamepad::end()
{
}
// 設定按鍵
void BleGamepad::setButton(uint8_t button, uint8_t value)
{
	if (value == 0)
	{
		releaseButton(button);
	}
	else
	{
		pressButton(button);
	}
}
// 按下按鍵
void BleGamepad::pressButton(uint8_t button)
{
	bitSet(buttons, button);
	if (autoSendState)
		sendState();
}
// 放開按鍵
void BleGamepad::releaseButton(uint8_t button)
{
	bitClear(buttons, button);
	if (autoSendState)
		sendState();
}
// 設定類比搖桿X軸（左搖桿X）
void BleGamepad::setXAxis(uint8_t value)
{
	xAxis = value;
	if (autoSendState)
		sendState();
}
// 設定類比搖桿Y軸（左搖桿Y）
void BleGamepad::setYAxis(uint8_t value)
{
	yAxis = value;
	if (autoSendState)
		sendState();
}
// 右搖桿X
void BleGamepad::setZAxis(uint8_t value)
{
	zAxis = value;
	if (autoSendState)
		sendState();
}
// 右搖桿Y
void BleGamepad::setZAxisRotation(uint8_t value)
{
	zAxisRotation = value;
	if (autoSendState)
		sendState();
}
// 設定十字鍵
void BleGamepad::setHatSwitch(int16_t value)
{
	hatSwitch = value;
	if (autoSendState)
		sendState();
}
// 傳送HID報告
void BleGamepad::sendState()
{
	if (this->isConnected())
	{
		uint8_t data[GAMEPAD_STATE_SIZE];
		uint32_t buttonTmp = buttons;

		// 把16位元拆分成2位元組
		data[0] = buttonTmp & 0xFF;
		buttonTmp >>= 8;
		data[1] = buttonTmp & 0xFF;

		if (hatSwitch < 0)
			hatSwitch = 8;

		// 把十字鍵按鍵狀態包裝成一個位元組
		data[2] = (B00001111 & hatSwitch);
		data[3] = xAxis;
		data[4] = yAxis;
		data[5] = zAxis;
		data[6] = zAxisRotation;

		this->inputGamepad->setValue(data, GAMEPAD_STATE_SIZE);
		this->inputGamepad->notify();
	}
}
// 確認藍牙的連線狀態
bool BleGamepad::isConnected()
{
	return this->connectionStatus->connected;
}
// 設定電量準位
void BleGamepad::setBatteryLevel(uint8_t level)
{
	this->batteryLevel = level;
	if (hid != 0)
		this->hid->setBatteryLevel(this->batteryLevel);
}